"""Utilities for testing future SNAP Connect"""
import os
import binascii
from itertools import cycle
import traceback
import gc

import mock
from tornado.testing import AsyncTestCase
from tornado.gen import coroutine, moment, Future

from snapconnect import snap

from snapconnect_futures import SnapConnectFutures


class SnapConnectFuturesTestCase(AsyncTestCase):
    """Test that the SNAP Connect futures wrapper works correctly"""

    def setUp(self):
        super(SnapConnectFuturesTestCase, self).setUp()
        self.fsc = self.setUpFsc()
        self.snapconnect = self.fsc._snap
        self.local_addr = self.snapconnect.local_addr()

        # Remove randomness from tests
        self.fsc._app_seq = 0

        assert isinstance(self.fsc, SnapConnectFutures)

    def tearDown(self):
        # Check for any outstanding futures
        uncollected_futures = [obj for obj in gc.get_objects() if isinstance(obj, Future)]

        for future in uncollected_futures:

            if future._log_traceback:  # pragma: no cover
                # set_exception() was called, or result() or exception() were not called

                tb = traceback.format_exception(*future._exc_info)
                print tb
                self.fail('Future exception for %s was never retrieved: %s' % (future, ''.join(tb).rstrip()))

        super(SnapConnectFuturesTestCase, self).tearDown()

    def setUpFsc(self):
        """Setup the FSC instance for the test

        This may be overriden by subclasses to alter parts of the setup if desired.

        This function MUST return a valid fsc object
        """
        snapconnect = get_mock_snapcom()
        return SnapConnectFutures(snapconnect)

    def assertRpcCalls(self, mock_calls):
        """Asert that a given list of mock calls was made to RPC"""
        self.assertEqual(self.snapconnect.rpc.mock_calls, mock_calls)
        self.snapconnect.rpc.reset_mock()

    def assertMcastCalls(self, mock_calls):
        """Asert that a given list of mock calls was made via mcast"""
        self.assertEqual(self.snapconnect.mcast_rpc.mock_calls, mock_calls)
        self.snapconnect.mcast_rpc.reset_mock()

    def assertDMcastCalls(self, mock_calls):
        """Asert that a given list of mock calls was made via dmcast"""
        self.assertEqual(self.snapconnect.dmcast_rpc.mock_calls, mock_calls)
        self.snapconnect.dmcast_rpc.reset_mock()


@coroutine
def moments():
    """Burn through all coroutines on the ioloop (Not timeouts)"""
    # TODO - Is there something in the ioloop we can look at to know the real number?
    for x in xrange(10):
        yield moment


def get_mock_snapcom(mock_scheduler=None):
    """Return a mock SNAP Connect object"""
    nv_param_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'temp', "%s.dat" % __name__)

    license_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'snapconnect_futures', "TestLicense.dat")

    snapcom = snap.Snap(nvparams_file=nv_param_path,
                        license_file=license_path,
                        funcs={},
                        scheduler=mock_scheduler)

    def mock_rx(addr, name, *args):
        """Mock the receipt of an RPC from a node

        :param addr: Hexlified SNAP Address (Ex: 'aabbcc')
        :param name: Name of RPC to call
        :param *args: Arguments for the RPC
        """
        prev_return_value = snapcom.rpc_source_addr.return_value
        snapcom.packet_dispatcher.remote_addr = binascii.unhexlify(addr)
        snapcom.rpc_source_addr.return_value = binascii.unhexlify(addr)

        # TODO - Re-enable these later - Right now the fingerprint tests break because they fake out the script push
        # descriptor = RpcCodec.RpcDescriptor()
        # descriptor.set_srcAddr(binascii.unhexlify(addr))
        # descriptor.funcName = name
        # descriptor.args = args
        # snapcom.local_dispatcher.on_rpc(descriptor)
        snapcom.local_dispatcher.funcs[name](*args)

        snapcom.packet_dispatcher.remote_addr = None
        snapcom.rpc_source_addr.return_value = prev_return_value

    snapcom.mock_rx = mock_rx

    def mock_rpc_sent(sequence, success):
        snapcom.packet_dispatcher._notify_callbacks(snap.hooks.HOOK_RPC_SENT, sequence, success)

    snapcom.mock_rpc_sent = mock_rpc_sent

    packet_sequence = cycle(xrange(255))
    snapcom.rpc = mock.Mock()
    snapcom.rpc.side_effect = packet_sequence
    snapcom.dmcast_rpc = mock.Mock()
    snapcom.dmcast_rpc.side_effect = packet_sequence
    snapcom.mcast_rpc = mock.Mock()
    snapcom.mcast_rpc.side_effect = packet_sequence
    snapcom.rpc_source_addr = mock.Mock()
    snapcom.rpc_source_addr.return_value = '\xaa\xbb\xcc'

    snapcom.open_serial = mock.Mock()
    snapcom.close_serial = mock.Mock()

    return snapcom
